Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Value.BuildSourceConvenience constructors that use zero_loc as the source location.
These are useful when building value trees programmatically and there's no meaningful source location or the tree is destined to be written as a YAML document straight away.
module Y = YAMLx.Value.Build
let v =
Y.map [
"name", Y.string "Alice";
"scores", Y.seq [ Y.int 95; Y.int 87 ]
]or
let v =
let open YAMLx.Value.Build in
map [
"name", string "Alice";
"scores", seq [ int 95; int 87 ]
]Int (zero_loc, Int64.of_int i). Use this for ordinary OCaml ints. See also int64.